"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const core_contracts_1 = require("@essential-projects/core_contracts");
const calculated_property_1 = require("./calculated_property");
const property_1 = require("./property");
class PropertyBag {
    constructor(datastoreServiceFactory, metadataProvider, context) {
        this._datastoreServiceFactory = undefined;
        this._metadataProvider = undefined;
        this._context = undefined;
        this._properties = {};
        this._propertyChangedCallbacks = {};
        this.isNew = true;
        this._datastoreServiceFactory = datastoreServiceFactory;
        this._metadataProvider = metadataProvider;
        this._context = context;
    }
    get datastoreServiceFactory() {
        return this._datastoreServiceFactory;
    }
    get metadataProvider() {
        return this._metadataProvider;
    }
    get context() {
        return this._context;
    }
    get properties() {
        return this._properties;
    }
    get propertyChangedCallbacks() {
        return this._propertyChangedCallbacks;
    }
    get propertyKeys() {
        return Object.keys(this.properties);
    }
    get dirtyPropertyKeys() {
        return this.propertyKeys.filter((propertyKey) => {
            const property = this._getProperty(propertyKey);
            return property.isDirty();
        });
    }
    async setupPropertyBag() {
        const initializeValuePromises = this.propertyKeys.map(async (propertyKey) => {
            const property = this._getProperty(propertyKey);
            await property.initializeValue();
        });
        await Promise.all(initializeValuePromises);
    }
    async _pojofyEntityCollectionProperty(context, pojoRef, propertyKey, value, options) {
        let relPojoOptions = Object.assign({}, options);
        relPojoOptions._currentDepth = relPojoOptions._currentDepth + 1;
        if (options && options.mode && options.mode === core_contracts_1.SaveMode.persistance) {
            relPojoOptions = {
                _currentDepth: relPojoOptions._currentDepth,
                maxDepth: relPojoOptions.maxDepth,
            };
        }
        return pojoRef[propertyKey] = await value.toPojos(context, relPojoOptions);
    }
    async _pojofyResolvedEntityProperty(context, pojoRef, propertyKey, value, options) {
        let relPojoOptions = Object.assign({}, options);
        delete relPojoOptions.select;
        relPojoOptions._currentDepth = relPojoOptions._currentDepth + 1;
        if (options && options.mode && options.mode === core_contracts_1.SaveMode.persistance) {
            relPojoOptions = {
                _currentDepth: relPojoOptions._currentDepth,
                maxDepth: relPojoOptions.maxDepth,
            };
        }
        pojoRef[propertyKey] = await value.toPojo(context, relPojoOptions);
    }
    async _pojofyEntityProperty(context, pojoRef, property, propertyKey, value, options) {
        const isEntityReference = property.isEntityReference();
        const isEntity = property.isEntity();
        if (isEntity && !isEntityReference) {
            await this._pojofyResolvedEntityProperty(context, pojoRef, propertyKey, value, options);
        }
        else if (!isEntity && !isEntityReference) {
            pojoRef[propertyKey] = value;
        }
        else if (isEntityReference) {
            pojoRef[propertyKey] = value.toPojo();
        }
    }
    async _pojofyComplexType(context, pojoRef, property, propertyKey, value, options) {
        const isEntityCollectionReference = property.isEntityCollectionReference();
        if ((isEntityCollectionReference || value && value.constructor && value.constructor.name === 'EntityCollection')) {
            if (!isEntityCollectionReference) {
                await this._pojofyEntityCollectionProperty(context, pojoRef, propertyKey, value, options);
            }
            else {
                pojoRef[propertyKey] = value.toPojo();
            }
        }
        else {
            await this._pojofyEntityProperty(context, pojoRef, property, propertyKey, value, options);
        }
    }
    async _exportProperty(context, pojoRef, propertyThisContext, propertyKey, options) {
        const bag = this;
        const property = bag._getProperty(propertyKey);
        if (property !== undefined) {
            const isComplexType = property.schema.isComplexType;
            const isCalculated = !!property.schema.calculated;
            let value;
            if (isCalculated && (!options || !options.skipCalculation)) {
                value = await bag.getCalculatedProperty(propertyThisContext, propertyKey, context);
            }
            else {
                value = bag.getProperty(propertyThisContext, propertyKey);
            }
            if (value !== null && value !== undefined) {
                if (isComplexType) {
                    await this._pojofyComplexType(context, pojoRef, property, propertyKey, value, options);
                }
                else {
                    pojoRef[propertyKey] = value;
                }
            }
            else {
                if (value !== undefined) {
                    pojoRef[propertyKey] = value;
                }
            }
        }
    }
    async toPojo(propertyThisContext, context, options, propertyKeys) {
        const propertyKeysToExport = propertyKeys || this.propertyKeys;
        const result = {};
        const promises = propertyKeysToExport
            .map((propertyKey) => this._exportProperty(context, result, propertyThisContext, propertyKey, options));
        return Promise.all(promises)
            .then(() => result)
            .catch((err) => {
            throw new Error(err);
        });
    }
    isDirty() {
        const isDirty = this.propertyKeys.some((propertyKey) => {
            const property = this._getProperty(propertyKey);
            return property.isDirty();
        });
        return isDirty;
    }
    _getProperty(propertyName) {
        return this.properties[propertyName];
    }
    _setProperty(propertyName, property) {
        this.properties[propertyName] = property;
    }
    getProperty(propertyThisContext, propertyName) {
        const property = this._getProperty(propertyName);
        const value = property.getValue(propertyThisContext);
        return value;
    }
    async getPropertyLazy(propertyThisContext, propertyName, context, options) {
        const property = this._getProperty(propertyName);
        const value = await property.getValueLazy(propertyThisContext, context, options);
        return value;
    }
    setProperty(propertyThisContext, propertyName, value) {
        const property = this._getProperty(propertyName);
        property.setValue(propertyThisContext, value);
    }
    async getCalculatedProperty(propertyThisContext, propertyName, context) {
        const property = this._getProperty(propertyName);
        const value = await property.getValueLazy(propertyThisContext, context);
        return value;
    }
    async setCalculatedProperty(propertyThisContext, propertyName, value, context) {
        const property = this._getProperty(propertyName);
        property.setValueLazy(propertyThisContext, value, context);
    }
    createProperty(schema, propertyName, onInitCallback) {
        const property = new property_1.Property(schema, this.datastoreServiceFactory, this.context, onInitCallback);
        this._setProperty(propertyName, property);
    }
    createCalculatedProperty(schema, propertyName, getCallback, setCallback) {
        let property = this._getProperty(propertyName);
        if (property) {
            throw new Error(`property for name ${propertyName} is already declared.`);
        }
        property = new calculated_property_1.CalculatedProperty(schema, this.datastoreServiceFactory, this.context, propertyName, getCallback, setCallback);
        this._setProperty(propertyName, property);
    }
    updateFromLocal(data, propertyThisContext) {
        this._setPropertyValues(data, false, propertyThisContext);
    }
    updateFromRemote(data) {
        this.isNew = false;
        this._setPropertyValues(data, true);
    }
    _setPropertyValues(data, fromRemote, propertyThisContext) {
        const propertyKeys = Object.keys(data);
        propertyKeys.map((propertyKey) => {
            const property = this._getProperty(propertyKey);
            if (property) {
                const value = data[propertyKey];
                if (fromRemote) {
                    property.updateFromRemote(value);
                }
                else {
                    property.setValue(propertyThisContext, value);
                }
            }
        });
    }
}
exports.PropertyBag = PropertyBag;

//# sourceMappingURL=property_bag.js.map
