import {
  ExecutionContext, IEntity, IEntityCollectionReference, IEntityReference, IExpand, IFromPojoOptions, IInheritedSchema, IPojoMetadata,
  IPrivateQueryOptions, IPrivateRemoveOptions, IPrivateSaveOptions, ISchemaAttribute, IToPojoOptions,
  SaveMode, SchemaAttributeType, TypeKey,
} from '@essential-projects/core_contracts';
import {IInvoker} from '@essential-projects/invocation_contracts';
import {schemaAttribute} from '@essential-projects/metadata';
import {IDecryptOptions, IEncryptOptions, ISecurityService} from '@essential-projects/security_service_contracts';

import {EntityCollection, EntityCollectionReference, EntityReference} from './index';
import {IDatastoreService, IEntityCollection, IEntityDependencyHelper, IEntityType, IProperty, IPropertyBag} from './index';

import * as toposort from 'toposort';
import * as uuidModule from 'uuid';
import {v4} from 'uuid/interfaces';

const uuid: any = uuidModule;

export class Entity implements IEntity {

  private _entityDependencyHelper: IEntityDependencyHelper = undefined;
  private _propertyBag: IPropertyBag = undefined;
  private _derivedClassInstance: IEntity = undefined;
  private _context: ExecutionContext = undefined;
  private _schema: IInheritedSchema = undefined;
  private _options: IFromPojoOptions = undefined;
  private _entityType: IEntityType<IEntity> = undefined;

  constructor(entityDependencyHelper: IEntityDependencyHelper,
              context: ExecutionContext,
              schema: IInheritedSchema,
              propertyBag: IPropertyBag,
              entityType: IEntityType<IEntity>) {
    this._entityDependencyHelper = entityDependencyHelper;
    this._propertyBag = propertyBag;
    this._context = context;
    this._schema = schema;
    this._entityType = entityType;
  }

  public get context(): ExecutionContext {
    return this._context;
  }

  public get schema(): IInheritedSchema {
    return this._schema;
  }

  protected get invoker(): IInvoker {
    return this._entityDependencyHelper.invoker;
  }

  protected get entityType(): IEntityType<IEntity> {
    return this._entityType;
  }

  protected get securityService(): ISecurityService {
    return this._entityDependencyHelper.securityService;
  }

  protected get propertyBag(): IPropertyBag {
    return this._propertyBag;
  }

  protected async getDatastoreService(): Promise<IDatastoreService> {
    return this._entityDependencyHelper.getDatastoreService();
  }

  public get isNew(): boolean {
    return this.propertyBag.isNew;
  }

  public get isDirty(): boolean {
    return this.propertyBag.isDirty();
  }

  private get derivedClassInstance(): IEntity {
    return this._derivedClassInstance;
  }

  public async initialize(derivedClassInstance: IEntity): Promise<void> {
    if (!derivedClassInstance) {
      throw new Error('derived class instance is missing');
    }
    this._derivedClassInstance = derivedClassInstance;
    this._propertyBag = await this.createPropertyBag(this.schema);
  }

  private async createPropertyBag(schema: IInheritedSchema): Promise<IPropertyBag> {

    const propertyBag: IPropertyBag = this.propertyBag;

    const attributeKeys: Array<string> = Object.keys(schema.attributes);

    attributeKeys.forEach((attributeKey: string) => {

      const attribute: ISchemaAttribute = schema.attributes[attributeKey];

      if (attribute.calculated) {

        const getCallback: Function = attribute.calculated.onGet;
        const setCallback: Function = attribute.calculated.onSet;

        propertyBag.createCalculatedProperty(attribute, attributeKey, getCallback, setCallback);

      } else {

        const onInitCallback: Function = attribute.onInit;

        propertyBag.createProperty(attribute, attributeKey, onInitCallback);
      }
    });

    await propertyBag.setupPropertyBag();

    return propertyBag;
  }

  @schemaAttribute({
    type: SchemaAttributeType.string,
    isPrimaryKey: true,
    onInit: (): v4 => {
      return uuid.v4();
    },
  })
  public get id(): string {
    return this.getProperty(this, 'id');
  }

  public set id(value: string) {
    this.setProperty(this, 'id', value);
  }

  public getProperty(propertyThisContext: any, propertyName: string): any {
    return this.propertyBag.getProperty(propertyThisContext, propertyName);
  }

  public async getPropertyLazy(propertyThisContext: any,
                               propertyName: string,
                               context: ExecutionContext,
                               options?: IPrivateQueryOptions): Promise<any> {
    return this.propertyBag.getPropertyLazy(propertyThisContext, propertyName, context, options);
  }

  public setProperty(propertyThisContext: any, propertyName: string, value: any): void {
    const propertySchema: ISchemaAttribute = this.schema.attributes ? this.schema.attributes[propertyName] : undefined;
    const propertyIsValid: boolean = propertySchema !== undefined && value !== undefined;
    if (!propertyIsValid) {
      return this.propertyBag.setProperty(propertyThisContext, propertyName, value);
    }

    const isPojo: boolean = propertySchema.isComplexType && !this._isResolvableEntity(value);
    if (!isPojo) {
      return this.propertyBag.setProperty(propertyThisContext, propertyName, value);
    }

    const valueIsValidObject: boolean = typeof value === 'object' && value.id !== undefined;
    const valueIsString: boolean = typeof value === 'string';
    const canResolvePropertyToEntity: boolean = valueIsValidObject || valueIsString;
    if (!canResolvePropertyToEntity) {
      throw new Error(`unable to construct a resolvable entity representation for property ${propertyName} with data ${JSON.stringify(value)}!`);
    }

    let resolvedValue: EntityReference;

    if (valueIsValidObject) {
      resolvedValue = new EntityReference(this.schema.namespace, propertySchema.type as TypeKey, value.id);
    } else if (valueIsString) {
      resolvedValue = new EntityReference(this.schema.namespace, propertySchema.type as TypeKey, value);
    }

    return this.propertyBag.setProperty(propertyThisContext, propertyName, resolvedValue);
  }

  public getCalculatedProperty(propertyThisContext: any, propertyName: string, context: ExecutionContext): Promise<any> {
    return this.propertyBag.getCalculatedProperty(propertyThisContext, propertyName, context);
  }

  public setCalculatedProperty(propertyThisContext: any, propertyName: string, value: any, context: ExecutionContext): Promise<void> {
    return this.propertyBag.setCalculatedProperty(propertyThisContext, propertyName, value, context);
  }

  public async toPojo(context: ExecutionContext, options?: IToPojoOptions): Promise<any> {

    const persistanceMode: boolean = options && options.mode && options.mode === SaveMode.persistance;
    const select: Array<string> = options && options.select ? options.select : undefined;

    options = options || {};
    if (!options.hasOwnProperty('_currentDepth')) {
      options._currentDepth = 0;
    }

    if (persistanceMode) {
      return this._toPojoForPersistance(context, options);
    }

    const currentDepth: number = options._currentDepth;

    // tslint:disable:no-magic-numbers
    const maxDepth: number = options.hasOwnProperty('maxDepth') ? options.maxDepth : 10;

    const datastoreService: IDatastoreService = await this.getDatastoreService();
    const sealInternals: boolean = (datastoreService && datastoreService.config && datastoreService.config.sealInternals);
    const hideInternals: boolean = sealInternals || (options && options.hideInternals);

    let data: any;

    if (maxDepth === undefined || currentDepth <= maxDepth) {

      let propertyKeysToExport: Array<string> = this.propertyBag.propertyKeys;

      if (select) {
        propertyKeysToExport = this._filterPropertyKeysBySelect(propertyKeysToExport, select);
      }

      if (hideInternals) {
        propertyKeysToExport = this._filterPropertyKeysByInternal(propertyKeysToExport);
      }

      data = await this.propertyBag.toPojo(this.derivedClassInstance, context, options, propertyKeysToExport);

      if (!hideInternals) {

        const meta: IPojoMetadata = {
          type: this.entityType.type,
          namespace: this.entityType.namespace,
          isRef: false,
          isNew: this.isNew,
        };

        data._meta = meta;
      }
    } else {
      data = this.getEntityReference().toPojo();
    }

    return data;
  }

  private async _toPojoForPersistance(context: ExecutionContext, options?: IToPojoOptions): Promise<any> {

    const saveFullEntity: boolean = options && options.saveFullEntity;

    const propertyKeysToExport: Array<string> = saveFullEntity ? this.propertyBag.propertyKeys : this.propertyBag.dirtyPropertyKeys;

    options = options || {};
    options.maxDepth = 0;

    const data: any = await this.propertyBag.toPojo(this.derivedClassInstance, context, options, propertyKeysToExport);

    await this._encryptData(data);

    return data;
  }

  private _filterPropertyKeysBySelect(propertyKeys: Array<string>, select: Array<string>): Array<string> {
    return select.filter((propertyKey: string) => {
      return (propertyKeys.indexOf(propertyKey) !== -1);
    });
  }

  private _filterPropertyKeysByInternal(propertyKeys: Array<string>): Array<string> {

    return propertyKeys.filter((propertyKey: string) => {

      const attributeSchema: ISchemaAttribute = this.schema.attributes[propertyKey];

      return attributeSchema && !attributeSchema.isInternal;
    });
  }

  public async fromPojo(context: ExecutionContext, data: any, options?: IFromPojoOptions): Promise<void> {

    if (!data) {
      return;
    }

    const dataComesFromAdapter: boolean = options && options.mode && options.mode === SaveMode.persistance;

    const propertyKeys: Array<string> = Object.keys(data);

    await this._resolveEntityReferences(data, propertyKeys);

    if (dataComesFromAdapter) {

      await this._fromRemotePojo(context, data, propertyKeys);

      await this._resolveEntityInstances(data, propertyKeys, options);

      await this._resolveEntityCollections(context, propertyKeys, options);

      await this.calculate(context, options);

    } else {

      this._filterAdapterMetadata(data, propertyKeys);

      await this._fromLocalPojo(context, data, propertyKeys);
    }

  }

  private async _resolveEntityCollections(context: ExecutionContext, propertyKeys: Array<string>, options?: IFromPojoOptions): Promise<void> {
    for (const propertyKey of propertyKeys) {
      const property: IProperty = this.propertyBag.properties[propertyKey];
      if (property.isEntityCollectionReference()) {
        const ref: any = property.getValue(null);
        if (ref && ref.resolve) {

          const queryOptions: IPrivateQueryOptions = {
            query: {
              attribute: ref.relatedAttribute,
              operator: '=',
              value: ref.id,
            },
          };
          const expandOption: Array<IExpand> = options.expandCollection || options.expandEntity;
          if (expandOption) {
            const foundExpands: Array<IExpand> = expandOption.filter((expand: IExpand) => propertyKey === expand.attribute);
            if (foundExpands.length === 1) {
              const expand: IExpand = foundExpands[0];
              if (expand.childAttributes && expand.childAttributes.length > 0) {
                queryOptions.select = expand.childAttributes.map((child: IExpand) => child.attribute);
              }
              queryOptions.expandCollection = expand.childAttributes;
            }
          }

          const dataStoreService: IDatastoreService = await this.getDatastoreService();
          const entityType: IEntityType<IEntity> = await dataStoreService.getEntityType(ref.relatedType);

          const coll: IEntityCollection<IEntity> = await entityType.query(context, queryOptions);

          this.propertyBag.setProperty(this, propertyKey, coll);

        }
      }
    }
  }

  public async calculate(context: ExecutionContext, options?: IFromPojoOptions): Promise<void> {
    const select: Array<string> = options && options.select ? options.select : undefined;

    let propertyKeys: Array<string> = this.propertyBag.propertyKeys;

    if (select) {
      propertyKeys = this._filterPropertyKeysBySelect(propertyKeys, select);
    }

    const topSortEdges: Array<Array<string>> = [];
    for (const propertyKey in propertyKeys) {
      const property: IProperty = this.propertyBag.properties[propertyKeys[propertyKey]];
      if (property.schema && property.schema.calculated) {
        if (property.schema.expandEntity) {
          property.schema.expandEntity.forEach((expand: IExpand) => {
            topSortEdges.push([propertyKeys[propertyKey], expand.attribute]);
          });
        } else {
          topSortEdges.push([propertyKeys[propertyKey], 'root']);
        }
      }
    }

    const topSortResult: Array<string> = toposort(topSortEdges).reverse().filter((edge: string) => edge !== 'root');

    for (const propertyKey in topSortResult) {
      const property: IProperty = this.propertyBag.properties[topSortResult[propertyKey]];
      if (property.schema && property.schema.calculated) {
        await property.getValueLazy(this, context);
      }
    }
  }

  private _isEntityReference(value: any): boolean {
    return value && value.constructor && value.constructor.name === 'EntityReference';
  }

  private _isEntityCollectionReference(value: any): boolean {
    return value && value.constructor && value.constructor.name === 'EntityCollectionReference';
  }

  private _isReference(value: any): boolean {
    return (this._isEntityCollectionReference(value) || this._isEntityReference(value));
  }

  private _isResolvableEntity(value: any): boolean {
    // if it looks like an Entity, it probably is one
    return this._isEntityReference(value) || value.getEntityReference !== undefined
                                          && value.entityType !== undefined
                                          && value.propertyBag !== undefined;
  }

  private _isResolvableEntityCollection(value: any): boolean {
    return this._isEntityCollectionReference(value) || value && value.constructor && value.constructor.name === 'EntityCollection';
  }

  private _isResolvable(value: any): boolean {
    return (this._isResolvableEntity(value) || this._isResolvableEntityCollection(value));
  }

  private _filterAdapterMetadata(data: any, propertyKeys: Array<string>): void {

    if (propertyKeys.indexOf('_meta') !== -1) {

      propertyKeys.splice(propertyKeys.indexOf('_meta'), 1);

      delete data._meta;
    }
  }

  private async _resolveEntityInstances(data: any, propertyKeys: Array<string>, options?: IFromPojoOptions): Promise<void> {

    const resolvePromises: Array<Promise<void>> = propertyKeys.map(async(propertyKey: string): Promise<void> => {

      const propertyValue: any = data[propertyKey];
      const propertySchema: ISchemaAttribute = this.schema.attributes ? this.schema.attributes[propertyKey] : undefined;

      if (propertySchema && propertySchema.isComplexType) {

        if (typeof propertyValue === 'undefined' || propertyValue === null || this._isReference(propertyValue)) {
          return;
        }

        const entityType: IEntityType<IEntity> = await (await this.getDatastoreService()).getEntityType(<string> propertySchema.type);

        const createEntityOption: IFromPojoOptions = Object.assign({}, options);
        if (createEntityOption.expandEntity) {
          const foundExpands: Array<IExpand> = createEntityOption.expandEntity.filter((expand: IExpand) => propertyKey === expand.attribute);
          if (foundExpands.length === 1) {
            const expand: IExpand = foundExpands[0];
            if (expand.childAttributes && expand.childAttributes.length > 0) {
              createEntityOption.select = expand.childAttributes.map((child: IExpand) => child.attribute);
            }
            createEntityOption.expandEntity = expand.childAttributes;
          }
        }
        const entity: IEntity = await entityType.createEntity(this.context, propertyValue, createEntityOption);

        data[propertyKey] = entity;
        this.propertyBag.setProperty(this, propertyKey, entity);
      }
    });

    await Promise.all(resolvePromises);
  }

  private async _resolveEntityReferences(data: any, propertyKeys: Array<string>): Promise<void> {

    const resolvePromises: Array<Promise<void>> = propertyKeys.map(async(propertyKey: string): Promise<void> => {

      const propertyValue: any = data[propertyKey];
      const propertySchema: ISchemaAttribute = this.schema.attributes ? this.schema.attributes[propertyKey] : undefined;

      const valueType: string = typeof propertyValue;

      if (valueType === 'undefined' || propertyValue === null) {
        return;
      }

      if (propertySchema && propertySchema.isComplexType) {

        if (propertySchema.isList) {
          const collectionReference: IEntityCollectionReference = new EntityCollectionReference(
            this.schema.namespace,
            propertySchema.type as string,
            propertyValue.id,
            propertySchema.type as string,
            propertySchema.relatedAttribute,
            propertyValue.resolve,
          );
          data[propertyKey] = collectionReference;

        } else if (valueType === 'string') {
          const entityReference: IEntityReference = new EntityReference(this.schema.namespace, propertySchema.type as TypeKey, propertyValue);

          data[propertyKey] = entityReference;
        } else {
          if (typeof propertyValue === 'object' && propertyValue._meta && propertyValue._meta.isRef && propertyValue.id) {
            const entityReference: IEntityReference = new EntityReference(this.schema.namespace, propertySchema.type as TypeKey, propertyValue.id);
            data[propertyKey] = entityReference;
          }
        }
      }
    });

    await Promise.all(resolvePromises);
  }

  private async _fromRemotePojo(context: ExecutionContext, data: any, propertyKeys: Array<string>): Promise<void> {

    await this._decryptData(data);

    this.propertyBag.updateFromRemote(data);
  }

  private async _fromLocalPojo(context: ExecutionContext, data: any, propertyKeys: Array<string>): Promise<void> {

    this.propertyBag.updateFromLocal(data, this);
  }

  private async _encryptData(data: any): Promise<void> {

    const keys: Array<string> = Object.keys(data);

    const encryptPromises: Array<Promise<void>> = keys.map(async(propertyKey: string): Promise<void> => {

      const attributeSchema: ISchemaAttribute = this.schema.attributes[propertyKey];
      const value: any = data[propertyKey];

      if (attributeSchema.encryptionType) {
        const encryptOptions: IEncryptOptions = <IEncryptOptions> {encryptionType: attributeSchema.encryptionType};
        const encryptedValue: string = await this.securityService.encrypt(value, encryptOptions);

        data[propertyKey] = encryptedValue;
      }
    });

    await Promise.all(encryptPromises);
  }

  private async _decryptData(data: any): Promise<void> {

    const keys: Array<string> = Object.keys(data);

    const decryptPromises: Array<Promise<void>> = keys.map(async(propertyKey: string): Promise<void> => {

      const attributeSchema: ISchemaAttribute = this.schema.attributes[propertyKey];
      const value: any = data[propertyKey];

      if (attributeSchema.encryptionType) {
        const decryptOptions: IDecryptOptions = <IDecryptOptions> {encryptionType: attributeSchema.encryptionType};
        const decryptedValue: any = await this.securityService.decrypt<any>(value, decryptOptions);

        data[propertyKey] = decryptedValue;
      }
    });

    await Promise.all(decryptPromises);
  }

  public async secureCompare(attributeName: string, valueToCompare: string): Promise<boolean> {
    const attributeSchema: ISchemaAttribute = this.schema.attributes[attributeName];
    const value: any = this[attributeName];

    const decryptOptions: IDecryptOptions = <IDecryptOptions> {encryptionType: attributeSchema.encryptionType};

    return await this.securityService.secureCompare(value, valueToCompare, decryptOptions);
  }

  public async save(context: ExecutionContext, options?: IPrivateSaveOptions): Promise<IEntity> {
    const reloadAfterSave: boolean = options && options.hasOwnProperty('reloadAfterSave') ? options.reloadAfterSave : true;
    const savedEntity: IEntity = await this.entityType.save(this, context, options);

    if (!reloadAfterSave) {
      this.propertyBag.isNew = false;
    }

    return savedEntity;
  }

  public async remove(context: ExecutionContext, options?: IPrivateRemoveOptions): Promise<boolean> {
    return await this.entityType.remove(this, context, options);
  }

  public getEntityReference(): IEntityReference {
    return new EntityReference(this.schema.namespace, this.entityType.type, this.id);
  }

}
