import {Entity} from './../entity';
import {EntityCollection} from './../entity_collection';
import {EntityCollectionReference} from './../entity_collection_reference';
import {EntityReference} from './../entity_reference';
import {ExpandType, IDatastoreService, IEntityCollection, IEntityType, IProperty} from './../index';

import {
  ExecutionContext, IEntity, IPrivateGetOptions, IPrivateQueryOptions,
  IQueryObject, ISchemaAttribute,
} from '@essential-projects/core_contracts';
import {IFactoryAsync} from 'addict-ioc';

import * as deepEqual from 'deep-equal';
import * as copy from 'deepcopy';

export class Property implements IProperty {

  private _schema: ISchemaAttribute = undefined;
  private _datastoreServiceFactory: IFactoryAsync<IDatastoreService> = undefined;
  private _datastoreService: IDatastoreService = undefined;
  private _context: ExecutionContext = undefined;
  private _onInitCallback: Function = undefined;
  private _remoteValue: any = undefined;
  private _value: any = undefined;

  constructor(schema: ISchemaAttribute,
              datastoreServiceFactory: IFactoryAsync<IDatastoreService>,
              context: ExecutionContext,
              onInitCallback: Function) {
    this._schema = schema;
    this._datastoreServiceFactory = datastoreServiceFactory;
    this._context = context;
    this._onInitCallback = onInitCallback;
  }

  public get schema(): ISchemaAttribute {
    return this._schema;
  }

  private async getDatastoreService(): Promise<IDatastoreService> {
    if (!this._datastoreService) {
      this._datastoreService = await this._datastoreServiceFactory();
    }

    return this._datastoreService;
  }

  private get context(): ExecutionContext {
    return this._context;
  }

  private get onInitCallback(): Function {
    return this._onInitCallback;
  }

  private get remoteValue(): any {
    return this._remoteValue;
  }

  private set remoteValue(value: any) {
    this._remoteValue = value;
  }

  private get value(): any {
    return this._value;
  }

  private set value(value: any) {
    this._value = value;
  }

  public isDirty(): boolean {
    // comparison can be made strict (===) by passing options with strict: true
    return !deepEqual(this.value, this.remoteValue);
  }

  public isEntity(): boolean {
    return this._isEntity(this.value);
  }

  public isEntityReference(): boolean {
    return this._isEntityReference(this.value);
  }

  public isEntityCollectionReference(): boolean {
    return this._isEntityCollectionReference(this.value);
  }

  public async initializeValue(): Promise<void> {

    if (this.onInitCallback) {

      const initialValue: Function = await this.onInitCallback();

      this.value = initialValue;
    }
  }

  public getValue(valueSource: any): any {
    return this._getCurrentValue();
  }

  private _buildEntityCollectionByReferenceQuery(currentValue: any, options?: IPrivateQueryOptions): IPrivateQueryOptions {

    const baseQuery: IQueryObject = {
      attribute: currentValue.relatedAttribute,
      operator: '=',
      value: currentValue.id,
    };

    const queryOptions: IPrivateQueryOptions = (options || {});

    // Compine a possible incoming query with baseQuery (relation)
    if (queryOptions.query) {
      queryOptions.query = {
        operator: 'and',
        queries: [
          queryOptions.query,
          baseQuery,
        ],
      };
    } else {
      queryOptions.query = baseQuery;
    }

    return queryOptions;
  }

  public async getValueLazy(valueSource: any, context: ExecutionContext, options?: IPrivateQueryOptions): Promise<any> {

    const currentValue: any = this._getCurrentValue();

    const isEntityReference: boolean = this._isEntityReference(currentValue);
    const isEntityCollectionReference: boolean = this._isEntityCollectionReference(currentValue);

    if (currentValue && currentValue.constructor && currentValue.constructor.name === 'EntityCollection') {
      return currentValue;
    }

    if (isEntityCollectionReference) {

      const queryOptions: IPrivateQueryOptions = this._buildEntityCollectionByReferenceQuery(currentValue, options);

      const coll: IEntityCollection<IEntity> = await this._getEntityCollectionByReference(currentValue.relatedType, context, queryOptions);

      this.remoteValue = this.value;
      this.value = coll;

      return coll;
    }

    if (!isEntityReference) {
      return currentValue;
    }

    let privateGetOptions: IPrivateGetOptions;

    if (isEntityReference) {
      const entityType: IEntityType<IEntity> = await (await this.getDatastoreService()).getEntityType(currentValue.type);
      privateGetOptions = {};
      privateGetOptions.expandEntity = entityType.getDefaultExpandOptions(ExpandType.Entity);
    } else {
      privateGetOptions = currentValue._options || {};
    }

    const entity: IEntity = await this._getEntityByReference(currentValue, context, privateGetOptions);

    this.remoteValue = this.value;
    this.value = entity;

    return entity;
  }

  public setValue(valueTarget: any, value: any): void {
    this._value = value;
  }

  public updateFromRemote(remoteValue: any): void {
    this._remoteValue = remoteValue;
    this._value = copy(remoteValue);
  }

  private _getCurrentValue(): any {
    return this.value === undefined ? this.remoteValue : this.value;
  }

  private _isEntity(value: any): boolean {
    // if it looks like an Entity, it probably is one
    return value.getEntityReference !== undefined && value.entityType !== undefined && value.propertyBag !== undefined;
  }

  private _isEntityReference(value: any): boolean {
    return value && value.constructor && value.constructor.name === 'EntityReference';
  }

  private _isEntityCollectionReference(value: any): boolean {
    return value && value.constructor && value.constructor.name === 'EntityCollectionReference';
  }

  private async _getEntityByReference(value: any, context: ExecutionContext, options?: IPrivateGetOptions): Promise<any> {

    const entityReference: EntityReference = value;

    const entityType: IEntityType<IEntity> = await (await this.getDatastoreService()).getEntityType(entityReference.type);

    const entity: IEntity = await entityType.getById(entityReference.id, context, options);

    return entity;
  }

  private async _getEntityCollectionByReference(entityCollectionTypeKey: string,
                                                context: ExecutionContext,
                                                options?: IPrivateQueryOptions): Promise<IEntityCollection<IEntity>> {

    const entityCollectionType: IEntityType<IEntity> = await (await this.getDatastoreService()).getEntityType(entityCollectionTypeKey);

    const coll: IEntityCollection<IEntity> = await entityCollectionType.query(context, options);

    return coll;
  }
}
