import {
  ExecutionContext, IPrivateQueryOptions, ISchemaAttribute,
  IToPojoOptions, SaveMode,
} from '@essential-projects/core_contracts';
import {IMetadataProvider} from '@essential-projects/metadata_contracts';
import {IFactoryAsync} from 'addict-ioc';
import {EntityCollection} from '../entity_collection';
import {ICalculatedProperty, IDatastoreService, IProperties, IProperty, IPropertyBag} from './../index';
import {CalculatedProperty} from './calculated_property';
import {Property} from './property';

interface IPropertyChangedCallbacks {
  [property: string]: Array<IPropertyChangedCallback>;
}

interface IPropertyChangedCallback {
  (property: string, value: any): void;
}

export class PropertyBag implements IPropertyBag {

  private _datastoreServiceFactory: IFactoryAsync<IDatastoreService> = undefined;
  private _metadataProvider: IMetadataProvider = undefined;
  private _context: ExecutionContext = undefined;
  private _properties: IProperties = {};
  private _propertyChangedCallbacks: IPropertyChangedCallbacks = {};
  public isNew: boolean = true;

  constructor(datastoreServiceFactory: IFactoryAsync<IDatastoreService>,
              metadataProvider: IMetadataProvider,
              context: ExecutionContext) {
    this._datastoreServiceFactory = datastoreServiceFactory;
    this._metadataProvider = metadataProvider;
    this._context = context;
  }

  private get datastoreServiceFactory(): IFactoryAsync<IDatastoreService> {
    return this._datastoreServiceFactory;
  }

  private get metadataProvider(): IMetadataProvider {
    return this._metadataProvider;
  }

  private get context(): ExecutionContext {
    return this._context;
  }

  public get properties(): IProperties {
    return this._properties;
  }

  private get propertyChangedCallbacks(): IPropertyChangedCallbacks {
    return this._propertyChangedCallbacks;
  }

  public get propertyKeys(): Array<string> {
    return Object.keys(this.properties);
  }

  public get dirtyPropertyKeys(): Array<string> {

    return this.propertyKeys.filter((propertyKey: string) => {

      const property: IProperty = this._getProperty(propertyKey);

      return property.isDirty();
    });
  }

  public async setupPropertyBag(): Promise<void> {

    const initializeValuePromises: Array<Promise<void>> = this.propertyKeys.map(async(propertyKey: string) => {

      const property: IProperty = this._getProperty(propertyKey);

      await property.initializeValue();
    });

    await Promise.all(initializeValuePromises);
  }

  private async _pojofyEntityCollectionProperty(context: ExecutionContext,
                                                pojoRef: any,
                                                propertyKey: string,
                                                value: any,
                                                options?: IToPojoOptions): Promise<any> {
    let relPojoOptions: IToPojoOptions = {...options};
    relPojoOptions._currentDepth = relPojoOptions._currentDepth + 1;
    if (options && options.mode && options.mode === SaveMode.persistance) {
      relPojoOptions = {
        _currentDepth: relPojoOptions._currentDepth,
        maxDepth: relPojoOptions.maxDepth,
      };
    }

    return pojoRef[propertyKey] = await value.toPojos(context, relPojoOptions);
  }

  private async _pojofyResolvedEntityProperty(context: ExecutionContext,
                                              pojoRef: any,
                                              propertyKey: string,
                                              value: any,
                                              options?: IToPojoOptions): Promise<void> {

    let relPojoOptions: IToPojoOptions = { ...options };
    delete relPojoOptions.select;
    relPojoOptions._currentDepth = relPojoOptions._currentDepth + 1;
    if (options && options.mode && options.mode === SaveMode.persistance) {
      relPojoOptions = {
        _currentDepth: relPojoOptions._currentDepth,
        maxDepth: relPojoOptions.maxDepth,
      };
    }

    pojoRef[propertyKey] = await value.toPojo(context, relPojoOptions);
  }

  private async _pojofyEntityProperty(context: ExecutionContext,
                                      pojoRef: any,
                                      property: IProperty,
                                      propertyKey: string,
                                      value: any,
                                      options?: IToPojoOptions): Promise<void> {

    const isEntityReference: boolean = property.isEntityReference();
    const isEntity: boolean = property.isEntity();

    if (isEntity && !isEntityReference) {
      await this._pojofyResolvedEntityProperty(context, pojoRef, propertyKey, value, options);
    } else if (!isEntity && !isEntityReference) {
      pojoRef[propertyKey] = value;
    } else if (isEntityReference) {
      pojoRef[propertyKey] = value.toPojo();
    }
  }

  private async _pojofyComplexType(context: ExecutionContext,
                                   pojoRef: any,
                                   property: IProperty,
                                   propertyKey: string,
                                   value: any,
                                   options?: IToPojoOptions): Promise<void> {

    const isEntityCollectionReference: boolean = property.isEntityCollectionReference();

    if ((isEntityCollectionReference || value && value.constructor && value.constructor.name === 'EntityCollection')) {
      if (!isEntityCollectionReference) {
        await this._pojofyEntityCollectionProperty(context, pojoRef, propertyKey, value, options);
      } else {
        pojoRef[propertyKey] = value.toPojo();
      }
    } else {
      await this._pojofyEntityProperty(context, pojoRef, property, propertyKey, value, options);
    }
  }

  private async _exportProperty(context: ExecutionContext,
                                pojoRef: any,
                                propertyThisContext: any,
                                propertyKey: string,
                                options?: IToPojoOptions): Promise<void> {

    const bag: PropertyBag = this;
    const property: IProperty = bag._getProperty(propertyKey);

    if (property !== undefined) {
      const isComplexType: boolean = property.schema.isComplexType;
      const isCalculated: boolean = !!property.schema.calculated;
      let value: any;

      if (isCalculated && (!options || !options.skipCalculation)) {
        value = await bag.getCalculatedProperty(propertyThisContext, propertyKey, context);
      } else {
        value = bag.getProperty(propertyThisContext, propertyKey);
      }

      if (value !== null && value !== undefined) {
        if (isComplexType) {
          await this._pojofyComplexType(context, pojoRef, property, propertyKey, value, options);
        } else {
          pojoRef[propertyKey] = value;
        }
      } else {
        if (value !== undefined) {
          pojoRef[propertyKey] = value;
        }
      }
    }
  }

  public async toPojo(propertyThisContext: any, context: ExecutionContext, options?: IToPojoOptions, propertyKeys?: Array<string>): Promise<any> {

    const propertyKeysToExport: Array<string> = propertyKeys || this.propertyKeys;

    const result: any = {};

    const promises: Array<Promise<void>> = propertyKeysToExport
      .map((propertyKey: string) => this._exportProperty(context, result, propertyThisContext, propertyKey, options));

    return Promise.all(promises)
      .then(() => result)
      .catch((err: string) => {
        throw new Error(err);
      });
  }

  public isDirty(): boolean {

    const isDirty: boolean = this.propertyKeys.some((propertyKey: string) => {

      const property: IProperty = this._getProperty(propertyKey);

      return property.isDirty();
    });

    return isDirty;
  }

  private _getProperty(propertyName: string): IProperty | ICalculatedProperty {
    return this.properties[propertyName];
  }

  private _setProperty(propertyName: string, property: IProperty): void {
    this.properties[propertyName] = property;
  }

  public getProperty(propertyThisContext: any, propertyName: string): any {

    const property: IProperty = this._getProperty(propertyName);

    const value: any = property.getValue(propertyThisContext);

    return value;
  }

  public async getPropertyLazy(propertyThisContext: any,
                               propertyName: string,
                               context: ExecutionContext,
                               options?: IPrivateQueryOptions): Promise<any> {

    const property: IProperty = this._getProperty(propertyName);

    const value: any = await property.getValueLazy(propertyThisContext, context, options);

    return value;
  }

  public setProperty(propertyThisContext: any, propertyName: string, value: any): void {

    const property: IProperty = this._getProperty(propertyName);

    property.setValue(propertyThisContext, value);
  }

  public async getCalculatedProperty(propertyThisContext: any, propertyName: string, context: ExecutionContext): Promise<any> {

    const property: ICalculatedProperty = <ICalculatedProperty> this._getProperty(propertyName);

    const value: any = await property.getValueLazy(propertyThisContext, context);

    return value;
  }

  public async setCalculatedProperty(propertyThisContext: any, propertyName: string, value: any, context: ExecutionContext): Promise<void> {

    const property: ICalculatedProperty = <ICalculatedProperty> this._getProperty(propertyName);

    property.setValueLazy(propertyThisContext, value, context);
  }

  public createProperty(schema: ISchemaAttribute, propertyName: string, onInitCallback: Function): void {

    const property: IProperty = new Property(schema, this.datastoreServiceFactory, this.context, onInitCallback);

    this._setProperty(propertyName, property);
  }

  public createCalculatedProperty(schema: ISchemaAttribute, propertyName: string, getCallback: Function, setCallback: Function): void {

    let property: IProperty = this._getProperty(propertyName);

    if (property) {
      throw new Error(`property for name ${propertyName} is already declared.`);
    }

    property = new CalculatedProperty(schema, this.datastoreServiceFactory, this.context, propertyName, getCallback, setCallback);

    this._setProperty(propertyName, property);
  }

  public updateFromLocal(data: any, propertyThisContext: any): void {

    this._setPropertyValues(data, false, propertyThisContext);
  }

  public updateFromRemote(data: any): void {

    this.isNew = false;

    this._setPropertyValues(data, true);
  }

  private _setPropertyValues(data: any, fromRemote: boolean, propertyThisContext?: any): void {

    const propertyKeys: Array<string> = Object.keys(data);

    propertyKeys.map((propertyKey: string) => {

      const property: IProperty = this._getProperty(propertyKey);

      if (property) {

        const value: any = data[propertyKey];

        if (fromRemote) {
          property.updateFromRemote(value);
        } else {
          property.setValue(propertyThisContext, value);
        }
      }

    });
  }

}
